"""Python library to make use of transport information from opendata.rmv.de."""
# pylint: disable=C0103
from .rmvtransport import RMVtransport  # noqa

MAJOR_VERSION = 0
MINOR_VERSION = 3
PATCH_VERSION = 1
__version__ = f"{MAJOR_VERSION}.{MINOR_VERSION}.{PATCH_VERSION}"



import pytest


@pytest.fixture
def stops_request():
    with open("fixtures/stops.response") as f:
        return f.read()


@pytest.fixture
def xml_request():
    with open("fixtures/request.xml") as f:
        return f.read()



"""This class represents a single journey."""
import html
import logging
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

from lxml import objectify  # type: ignore

from .const import IMG_URL, PRODUCTS

_LOGGER = logging.getLogger(__name__)


class RMVJourney:
    """A journey object to hold information about a journey."""

    # pylint: disable=I1101
    def __init__(self, journey: objectify.ObjectifiedElement, now: datetime) -> None:
        """Initialize the journey object."""
        self._journey: objectify.ObjectifiedElement = journey
        self._now: datetime = now
        self._attr_types = self._journey.JourneyAttributeList.xpath("*/Attribute/@type")

    def as_dict(self) -> Dict:
        """Build journey dictionary."""
        return {
            "product": self.product,
            "number": self.number,
            "trainId": self.train_id,
            "direction": self.direction,
            "departure_time": self.real_departure_time,
            "minutes": self.real_departure,
            "delay": self.delay,
            "stops": self.stops,
            "info": self.info,
            "info_long": self.info_long,
            "icon": self.icon,
        }

    @property
    def number(self) -> str:
        """Return the number of the route."""
        return self._extract("NUMBER")

    @property
    def product(self) -> str:
        """Return the product category."""
        return self._extract("CATEGORY")

    @property
    def train_id(self) -> str:
        """Return the train id."""
        return str(self._journey.get("trainId"))

    @property
    def departure(self) -> datetime:
        """Return time of departure."""
        return self._departure()

    @property
    def delay(self) -> int:
        """Return current delay for departure."""
        return self._delay()

    @property
    def real_departure_time(self) -> datetime:
        """Return the real departure time."""
        return self._real_departure_time()

    @property
    def real_departure(self) -> int:
        """Return minutes until departure."""
        return self._real_departure()

    @property
    def direction(self) -> str:
        """Return the direction of travel."""
        return self._extract("DIRECTION")

    @property
    def info(self) -> Optional[str]:
        """Return journey information."""
        return self._info()

    @property
    def info_long(self) -> Optional[str]:
        """Return long journey information."""
        return self._info_long()

    @property
    def _stops(self) -> List[Dict]:
        """Return list of stops along the journey."""
        return self._pass_list()

    @property
    def stops(self) -> List[str]:
        """Return list of stops along the journey."""
        return [s["station"] for s in self._stops]

    @property
    def icon(self) -> str:
        """Return icon url for the means of transport."""
        return self._icon()

    def _delay(self) -> int:
        """Extract departure delay."""
        try:
            return int(self._journey.MainStop.BasicStop.Dep.Delay.text)
        except AttributeError:
            return 0

    def _departure(self) -> datetime:
        """Extract departure time."""
        departure_time = datetime.strptime(
            self._journey.MainStop.BasicStop.Dep.Time.text, "%H:%M"
        ).time()
        if departure_time > (self._now - timedelta(hours=1)).time():
            return datetime.combine(self._now.date(), departure_time)
        return datetime.combine(self._now.date() + timedelta(days=1), departure_time)

    def _real_departure_time(self) -> datetime:
        """Calculate actual departure time."""
        return self.departure + timedelta(minutes=self.delay)

    def _real_departure(self) -> int:
        """Calculate actual minutes left for departure."""
        return round((self.real_departure_time - self._now).seconds / 60)

    def _extract(self, attribute) -> str:
        """Extract train information."""
        attr_data = self._journey.JourneyAttributeList.JourneyAttribute[
            self._attr_types.index(attribute)
        ].Attribute
        attr_variants = attr_data.xpath("AttributeVariant/@type")
        try:
            data = attr_data.AttributeVariant[attr_variants.index("NORMAL")].Text.pyval
        except ValueError:
            return ""
        return str(data)

    def _info(self) -> Optional[str]:
        """Extract journey information."""
        try:
            return str(html.unescape(self._journey.InfoTextList.InfoText.get("text")))
        except AttributeError:
            return None

    def _info_long(self) -> Optional[str]:
        """Extract journey information."""
        try:
            return str(
                html.unescape(self._journey.InfoTextList.InfoText.get("textL")).replace(
                    "<br />", "\n"
                )
            )
        except AttributeError:
            return None

    def _pass_list(self) -> List[Dict[str, Any]]:
        """Extract next stops along the journey."""
        stops: List[Dict[str, Any]] = []
        for stop in self._journey.PassList.BasicStop:
            index = stop.get("index")
            station = stop.Location.Station.HafasName.Text.text
            station_id = stop.Location.Station.ExternalId.text
            stops.append({"index": index, "stationId": station_id, "station": station})
        return stops

    def _icon(self) -> str:
        """Extract product icon."""
        pic_url = IMG_URL
        try:
            return pic_url % PRODUCTS[self.product]
        except KeyError:
            _LOGGER.debug("No matching icon for product: %s", self.product)
            return pic_url % PRODUCTS["Bahn"]



from datetime import datetime

URL = "www.rmv.de"
URL_PATH = "/auskunft/bin/jp/stboard.exe/dn"
URL_SEARCH_PATH = "/auskunft/bin/jp/ajax-getstop.exe/dn"


def date_hook(json_dict):
    """JSON datetime parser."""
    for (key, value) in json_dict.items():
        try:
            json_dict[key] = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
        except (TypeError, ValueError):
            pass
    return json_dict



"""Constants."""
from typing import List, Dict


PRODUCTS: Dict[str, int] = {
    "ICE": 1,
    "IC": 2,
    "EC": 2,
    "R": 4,
    "RB": 4,
    "RE": 4,
    "S": 8,
    "U-Bahn": 16,
    "Tram": 32,
    "Bus": 64,
    "Bus2": 128,
    "Fähre": 256,
    "Taxi": 512,
    "Bahn": 1024,
}

ALL_PRODUCTS: List[str] = list(PRODUCTS.keys())

MAX_RETRIES: int = 5

KNOWN_XML_ISSUES: Dict[str, str] = {"<Arr getIn=false>": "<Arr >"}

BASE_URI: str = "https://www.rmv.de/auskunft/"
PREFIX: str = "bin/jp/"
QUERY_PATH: str = PREFIX + "query.exe/"
GETSTOP_PATH: str = PREFIX + "ajax-getstop.exe/"
STBOARD_PATH: str = PREFIX + "stboard.exe/"
IMG_URL: str = BASE_URI + "s/n/img/products/%i_pic.png"



"""A module to query bus and train departure times."""
import asyncio
import json
import logging
import urllib.parse
import urllib.request
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import async_timeout
import httpx
from lxml import etree, objectify  # type: ignore

from .const import (
    ALL_PRODUCTS,
    BASE_URI,
    GETSTOP_PATH,
    KNOWN_XML_ISSUES,
    MAX_RETRIES,
    PRODUCTS,
    STBOARD_PATH,
)
from .errors import (
    RMVtransportApiConnectionError,
    RMVtransportDataError,
    RMVtransportError,
)
from .rmvjourney import RMVJourney

_LOGGER = logging.getLogger(__name__)


class RMVtransport:
    """Connection data and travel information."""

    def __init__(self, timeout: float = 10) -> None:
        """Initialize connection data."""
        self._timeout: float = timeout

        self.now: datetime

        self.station_id: str
        self.direction_id: Optional[str]
        self.products_filter: str

        self.max_journeys: int

        self.obj: objectify.ObjectifiedElement
        self.journeys: List[RMVJourney] = []

    async def get_departures(
        self,
        station_id: str,
        direction_id: Optional[str] = None,
        max_journeys: int = 20,
        products: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """Fetch data from rmv.de."""
        url = self.build_journey_query(station_id, direction_id, max_journeys, products)
        xml = await self._query_rmv_api(url)
        self.obj = extract_data_from_xml(xml)

        try:
            self.now = self.current_time()
        except RMVtransportDataError:
            _LOGGER.debug(
                "XML contains unexpected data %s", objectify.dump(self.obj)[:100]
            )
            raise

        self.journeys.clear()
        try:
            for journey in self.obj.SBRes.JourneyList.Journey:
                self.journeys.append(RMVJourney(journey, self.now))
        except AttributeError:
            _LOGGER.debug("Extract journeys: %s", objectify.dump(self.obj.SBRes))
            raise RMVtransportError()

        return self.travel_data()

    def build_journey_query(
        self,
        station_id: str,
        direction_id: Optional[str] = None,
        max_journeys: int = 20,
        products: Optional[List[str]] = None,
    ) -> str:
        """Build query to request journey data."""
        self.station_id = station_id
        self.direction_id = direction_id
        self.max_journeys = max_journeys
        self.products_filter = product_filter(products or ALL_PRODUCTS)

        params: Dict[str, Union[str, int]] = {
            "selectDate": "today",
            "time": "now",
            "input": self.station_id,
            "maxJourneys": self.max_journeys,
            "boardType": "dep",
            "productsFilter": self.products_filter,
            "disableEquivs": "discard_nearby",
            "output": "xml",
            "start": "yes",
        }
        if self.direction_id:
            params["dirInput"] = self.direction_id

        return base_url() + urllib.parse.urlencode(params)

    async def search_station(self, name: str, max_results: int = 25) -> Dict[str, Dict]:
        """Search station/stop my name."""
        suggestions: List = await self._fetch_sugestions(name, max_results)

        return {
            item["extId"]: {
                "id": item["extId"],
                "name": item["value"],
                "lat": convert_coordinates(item["ycoord"]),
                "long": convert_coordinates(item["xcoord"]),
            }
            for item in suggestions
        }

    async def _fetch_sugestions(
        self, name: str, max_results: int
    ) -> List[Optional[Dict]]:
        """Fetch suggestsions for the given station name from the backend."""
        params: Dict[str, Union[str, int]] = {
            "getstop": 1,
            "REQ0JourneyStopsS0A": max_results,
            "REQ0JourneyStopsS0G": name,
        }

        url = base_url(GETSTOP_PATH) + urllib.parse.urlencode(params)
        _LOGGER.debug("URL: %s", url)

        response = await self._query_rmv_api(url)
        data = extract_json_data(response)

        try:
            json_data = json.loads(data)
        except (TypeError, json.JSONDecodeError):
            _LOGGER.debug("Error in JSON: %s...", data[:100])
            raise RMVtransportError()

        return list(json_data["suggestions"][:max_results])

    async def _query_rmv_api(self, url: str) -> Any:
        """Query RMV API."""
        with async_timeout.timeout(self._timeout):
            async with httpx.AsyncClient() as client:
                try:
                    response = await client.get(url)
                except (
                    asyncio.TimeoutError,
                    httpx.ReadTimeout,
                    httpx.ConnectTimeout,
                    httpx.ConnectError,
                ):
                    _LOGGER.error("Can not load data from RMV API")
                    raise RMVtransportApiConnectionError()

        _LOGGER.debug("Response from RMV API: %s", response.status_code)
        return response.read()

    def travel_data(self) -> Dict[str, Any]:
        """Return travel data."""
        return {
            "station": self.station,
            "stationId": self.station_id,
            "filter": self.products_filter,
            "journeys": self.build_journey_list(),
        }

    def build_journey_list(self) -> List[Dict]:
        """Build list of journeys."""
        return [
            j.as_dict()
            for j in sorted(self.journeys, key=lambda k: k.real_departure)[
                : self.max_journeys
            ]
        ]

    @property
    def station(self) -> str:
        """Extract station name."""
        return str(self.obj.SBRes.SBReq.Start.Station.HafasName.Text.pyval)

    def current_time(self) -> datetime:
        """Extract current time."""
        try:
            _date = datetime.strptime(self.obj.SBRes.SBReq.StartT.get("date"), "%Y%m%d")
            _time = datetime.strptime(self.obj.SBRes.SBReq.StartT.get("time"), "%H:%M")
        except (ValueError, AttributeError):
            raise RMVtransportDataError()

        return datetime.combine(_date.date(), _time.time())

    def print(self) -> None:
        """Pretty print travel times."""
        result = [f"{self.station} - {self.now}"]

        for j in sorted(self.journeys, key=lambda k: k.real_departure)[
            : self.max_journeys
        ]:
            result.append("-------------")
            result.append(f"{j.product}: {j.number} ({j.train_id})")
            result.append(f"Richtung: {j.direction}")
            result.append(f"Abfahrt in {j.real_departure} min.")
            result.append(f"Abfahrt {j.departure.time()} (+{j.delay})")
            result.append(f"Nächste Haltestellen: {(j.stops)}")
            if j.info:
                result.append(f"Hinweis: {j.info}")
                result.append(f"Hinweis (lang): {j.info_long}")
            result.append(f"Icon: {j.icon}")

        print("\n".join(result))


def product_filter(products) -> str:
    """Calculate the product filter."""
    _filter = sum({PRODUCTS[p] for p in products})
    return format(_filter, "b")[::-1]


def base_url(path: str = STBOARD_PATH) -> str:
    """Build base url."""
    _lang: str = "d"
    _type: str = "n"
    _with_suggestions: str = "?"
    return BASE_URI + path + _lang + _type + _with_suggestions


def convert_coordinates(value: str) -> float:
    """Convert coordinates to lat/long."""
    if len(value) < 8:
        return float(value[0] + "." + value[1:])
    return float(value[0:2] + "." + value[2:])


def extract_data_from_xml(xml: bytes) -> Any:
    """Extract data from xml."""
    retry = 0
    while retry < MAX_RETRIES:
        try:
            return objectify.fromstring(xml)

        except etree.XMLSyntaxError as err:
            xml = fix_xml(xml, err)
            retry -= 1


def fix_xml(data: bytes, err: etree.XMLSyntaxError) -> Any:
    """Try to fix known issues in XML data."""
    xml_issue = data.decode().split("\n")[err.lineno - 1]

    if xml_issue not in KNOWN_XML_ISSUES.keys():
        _LOGGER.debug("Unknown xml issue in: %s", xml_issue)
        raise RMVtransportError()

    return data.decode().replace(xml_issue, KNOWN_XML_ISSUES[xml_issue]).encode()


def extract_json_data(response) -> str:
    """Extract json from response."""
    data = response.decode("utf-8")
    return str(data[data.find("{") : data.rfind("}") + 1])  # noqa: E203



"""Define package errors."""


class RMVtransportError(Exception):
    """General error exception occurred."""

    pass


class RMVtransportApiConnectionError(RMVtransportError):
    """When a connection error is encountered."""

    pass


class RMVtransportDataError(RMVtransportError):
    """When an error in the XML data is encountered."""

    pass



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['RMVtransport']

package_data = \
{'': ['*']}

install_requires = \
['async-timeout', 'lxml', 'httpx']

extras_require = \
{'test': ['pytest >= 5.2.2',
          'pytest-cov >= 2.8.1',
          'pytest-asyncio',
          'pylint',
          'tox',
          'pytest-mock',
          'pytest-httpx',
          'asynctest',
          'codecov',
          'coverage',
          'mock',
          'flake8',
          'mypy',
          'bandit',
          'pre-commit',
          'black']}

setup(name='PyRMVtransport',
      version='0.3.1',
      description='Python library to make use of transport information from opendata.rmv.de.',
      author='cgtobi',
      author_email='cgtobi@gmail.com',
      url='https://github.com/cgtobi/PyRMVtransport',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
